use anyhow::Result;
use axum::{
    http::StatusCode,
    response::IntoResponse,
    routing::get_service,
    Router,
};
use clap::{ArgAction, Parser};
use log::{error, info};
use std::{
    env, io,
    net::{IpAddr, SocketAddr},
    path::PathBuf,
};
use tower::ServiceBuilder;
use tower_http::{
    services::{ServeDir, ServeFile},
    trace::TraceLayer,
};

#[derive(Parser)]
#[clap(about, author, version)]
struct Cli {
    /// The address to listen on
    #[clap(short, long, default_value = "::1")]
    address: IpAddr,
    /// The port to listen on
    #[clap(short, long, default_value = "3000")]
    port: u16,
    /// Path to the directory containing the statically served files
    #[clap(short, long, default_value = "dist")]
    static_dir: String,
    /// Set the log level. Multiple -v options increase the verbosity
    #[clap(short, action = ArgAction::Count)]
    verbosity: u8,
}

#[allow(clippy::unused_async)]
async fn file_error(err: io::Error) -> impl IntoResponse {
    error!("could not read from file: {}", err);
    (StatusCode::INTERNAL_SERVER_ERROR, "Something went wrong...")
}

#[tokio::main]
async fn main() -> Result<()> {
    let cli = Cli::parse();

    if env::var("RUST_LOG").is_err() {
        env::set_var("RUST_LOG", format!("{},hyper=info,mio=info", cli.verbosity));
    }

    tracing_subscriber::fmt::init();

    let app = Router::new()
        .nest(
            "/client",
            get_service(ServeDir::new(&cli.static_dir)).handle_error(file_error),
        )
        .route(
            "/",
            get_service(ServeFile::new(
                &PathBuf::from(&cli.static_dir).join("index.html"),
            ))
            .handle_error(file_error),
        )
        .layer(ServiceBuilder::new().layer(TraceLayer::new_for_http()));

    let addr = SocketAddr::from((cli.address, cli.port));

    info!("listening on {}", addr);

    axum::Server::bind(&addr)
        .serve(app.into_make_service())
        .await?;

    Ok(())
}
